/* RTK 2008.2.4 */
#define n (1024 * 1024)
#define b 16
#define m ((1 << b) - 1)

typedef unsigned short BlokT[n];

/* Najbolj naivna reitev -- O(n b 2^b). */
unsigned short NajvecNicel1(const BlokT blok)
{
	int c, najC = -1, stNicel, najNicel = -1, i, j, x;
	for (c = 0; c <= m; c++)
	{
		/* Pojdimo po celem bloku in tejmo, koliko niel
		   ostane, e ga xoramo z vrednostjo c. */
		stNicel = 0;
		for (i = 0; i < n; i++)
			/* Po xoranju imamo v i-ti celici tevilo x; pretejmo nile v njem. */
			for (j = 0, x = blok[i] ^ c; j < b; j++)
				if (! (x & (1 << j))) stNicel++;
		if (stNicel > najNicel)	/* Najbolja reitev doslej. */
			najNicel = stNicel, najC = c;
	}
	return (unsigned short) najC;
}

/* Za vsak bit pretejemo, v koliko celicah bloka je prigan.  O(n b). */
unsigned short NajvecNicel3(const BlokT blok)
{
	unsigned short c; int i, j, stNicel[b];
	for (j = 0; j < b; j++) stNicel[j] = 0;
	/* Za vsak bit poglejmo, koliko niel je na tem mestu v naem bloku. */
	for (i = 0; i < n; i++)
		for (j = 0, c = blok[i]; j < b; j++)
			if (! (c & (1 << j))) stNicel[j]++;
	/* V c prigimo tiste bite, kjer je enic ve kot niel. */
	for (j = 0, c = 0; j < b; j++)
		if (stNicel[j] < n - stNicel[j]) c |= 1 << j;
	return c;
}

/* NajvecNicel3 + tabela s tevilom pojavitev za vsako mono vrednost.  O(n + b 2^b ). */
unsigned short NajvecNicel3b(const BlokT blok)
{
	unsigned short c; int i, j, f[m + 1], stNicel[b];
	/* Za vsako mono vrednost pretejmo, kolikokrat se pojavlja. */
	for (i = 0; i <= m; i++) f[i] = 0;
	for (i = 0; i < n; i++) f[blok[i]]++;
	/* Pretejmo, koliko je v bloku niel na vsakem bitu. */
	for (j = 0; j < b; j++) stNicel[j] = 0;
	for (i = 0; i <= m; i++) if (f[i] > 0)
		for (j = 0; j < b; j++) if (! (i & (1 << j))) stNicel[j] += f[i];
	/* V c prigimo tiste bite, kjer je enic ve kot niel. */
	for (j = 0, c = 0; j < b; j++)
		if (stNicel[j] < n - stNicel[j]) c |= 1 << j;
	return c;
}

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

double Ura()
{
	return (double) clock() / (double) CLOCKS_PER_SEC;
}

void InitBlok(BlokT blok)
{
	int i;
	for (i = 0; i < n; i++)
	{
		blok[i] = rand() % (1 << 10);
		blok[i] <<= 10;
		blok[i] |= rand() % (1 << 10);
		blok[i] &= m;
	}
}

void Test(unsigned short (*f)(const BlokT), const char *ime)
{
	BlokT *blok;
	double tm = 0, tmStart;
	int nIter = 0;
	unsigned short c;

	blok = (BlokT *) malloc(sizeof(BlokT));
	srand(1234);
	do
	{
		InitBlok(*blok);
		tmStart = Ura();
		c = f(*blok);
		tm += (Ura() - tmStart);
		if (++nIter == 1) printf("%s: c = %d", ime, c);
	}
	while (tm < 1.0);
	printf("; %.2f s for %d iterations = %.6f s/iter\n", tm, nIter, tm / nIter);
	free(blok);
}

int main()
{
	/*Test(NajvecNicel1, "NajvecNicel1");*/
	Test(NajvecNicel3, "NajvecNicel3");
	Test(NajvecNicel3b, "NajvecNicel3b");
	return 0;
}
