{$M 3000000,0}
program DrustvoLjubiteljevNicel; { RTK 2008.2.4 }
uses Windows;
const 
  n = 65536; {1024 * 1024;}
  b = 10;
  m = (1 shl b) - 1;
type 
  word = 0..m;
  BlokT = packed array [0..n - 1] of word;

{ Najbolj naivna reitev -- O(n b 2^b). }
function NajvecNicel1(const Blok: BlokT): word;
var
  c, NajC, StNicel, NajNicel, i, j, x: integer;
begin
  NajNicel := -1; NajC := -1;
  for c := 0 to m do begin 
    { Pojdimo po celem bloku in tejmo, koliko niel
      ostane, e ga xoramo z vrednostjo c. }
    StNicel := 0;
    for i := 0 to n - 1 do begin
      x := Blok[i] xor c;
      { Po xoranju imamo v i-ti celici tevilo x; pretejmo nile v njem. }
      for j := 0 to b - 1 do
        if not Odd(x shr j) then StNicel := StNicel + 1;
    end; {for i}
    if StNicel > NajNicel then { Najbolja reitev doslej. }
      begin NajNicel := StNicel; NajC := c end;
  end; {for c}
  NajvecNicel1 := word(NajC);
end; {NajvecNicel1}

{ NajvecNicel1 + za vse mone vrednosti si potabeliramo tevilo bitov. -- O(n 2^b). }
function NajvecNicel1b(const Blok: BlokT): word;
var
  StEnic: array [0..m] of integer;
  c, NajC, StNicel, NajNicel, i: integer;
begin
  { Pripravimo si tabelo s tevilom niel za vsako mono besedo. }
  StEnic[0] := 0;
  for c := 1 to m do StEnic[c] := StEnic[c shr 1] + (c and 1);
  { Preglejmo vse mone vrednosti c. }
  NajNicel := -1; NajC := -1;
  for c := 0 to m do begin 
    { Pojdimo po celem bloku in tejmo, koliko niel
      ostane, e ga xoramo z vrednostjo c. }
    StNicel := n shl b;
    for i := 0 to n - 1 do 
      StNicel := StNicel - StEnic[Blok[i] xor c];
    if StNicel > NajNicel then { Najbolja reitev doslej. }
      begin NajNicel := StNicel; NajC := c end;
  end; {for c}
  NajvecNicel1b := word(NajC);
end; {NajvecNicel1b}

{ Malo bolja reitev, e je n velik: ez cel blok gremo
  enkrat samkrat in za vsako mono vrednost pretejemo,
  kolikokrat se pojavlja v njem.  O(n + b 4^b). }
function NajvecNicel2(const Blok: BlokT): word;
var
  f: array [0..m] of integer; 
  c, NajC, d, StNicel, NajNicel, i, j, x, xn: integer;
begin
  { Za vsako mono vrednost pretejmo, kolikokrat se pojavlja. }
  for d := 0 to m do f[d] := 0;
  for i := 0 to n - 1 do f[Blok[i]] := f[Blok[i]] + 1;
  { Preizkusimo zdaj vse mone c. }
  NajNicel := -1; NajC := -1;
  for c := 0 to m do begin 
    { Preglejmo vse mone vrednosti in izraunajmo, koliko niel
      ostane v bloku po xoranju z vrednostjo c. }
    StNicel := 0;
    for d := 0 to m do if f[d] > 0 then begin
      x := d xor c; xn := 0;
      { Po xoranju iz d nastane tevilo x; pretejmo nile v njem. }
      for j := 0 to b - 1 do
        if not Odd(x shr j) then xn := xn + 1;
      { Upotevajmo, da d se pojavlja v bloku f[d]-krat. }
      StNicel := StNicel + xn * f[d];
    end; {for d}
    if StNicel > NajNicel then { Najbolja reitev doslej. }
      begin NajNicel := StNicel; NajC := c end;
  end; {for c}
  NajvecNicel2 := word(NajC);
end; {NajvecNicel2}

{ NajvecNicel2 + tabela za tevilo bitov v vsaki moni vrednosti. -- O(n + 4^b). }
function NajvecNicel2b(const Blok: BlokT): word;
var
  f, StEnic: array [0..m] of integer; 
  c, NajC, d, StNicel, NajNicel, i: integer;
begin
  { Pripravimo si tabelo s tevilom niel za vsako mono besedo. }
  StEnic[0] := 0;
  for d := 1 to m do StEnic[d] := StEnic[d shr 1] + (d and 1);
  { Za vsako mono vrednost pretejmo, kolikokrat se pojavlja. }
  for d := 0 to m do f[d] := 0;
  for i := 0 to n - 1 do f[Blok[i]] := f[Blok[i]] + 1;
  { Preizkusimo zdaj vse mone c. }
  NajNicel := -1; NajC := -1;
  for c := 0 to m do begin 
    { Preglejmo vse mone vrednosti in izraunajmo, koliko niel
      ostane v bloku po xoranju z vrednostjo c. }
    StNicel := n shl b;
    for d := 0 to m do 
      StNicel := StNicel - StEnic[d xor c] * f[d];
    if StNicel > NajNicel then { Najbolja reitev doslej. }
      begin NajNicel := StNicel; NajC := c end;
  end; {for c}
  NajvecNicel2b := word(NajC);
end; {NajvecNicel2b}

{ Za vsak bit pretejemo, v koliko celicah bloka je prigan.  O(n b). }
function NajvecNicel3(const Blok: BlokT): word;
var
  c: word; i, j: integer;
  StNicel: array [0..b - 1] of integer;
begin
  for j := 0 to b - 1 do StNicel[j] := 0;
  for i := 0 to n - 1 do begin
    c := Blok[i];
    for j := 0 to b - 1 do 
      if not Odd(c shr j) then StNicel[j] := StNicel[j] + 1;
  end; {for i}
  { V c prigimo tiste bite, kjer je enic ve kot niel. }
  c := 0;
  for j := 0 to b - 1 do
    if StNicel[j] < n - StNicel[j] then
      c := c or (1 shl j);
  NajvecNicel3 := c;
end; {NajvecNicel3}

{ NajvecNicel3 + tabela s tevilom pojavitev za vsako mono vrednost.  O(n + b 2^b ). }
function NajvecNicel3b(const Blok: BlokT): word;
var
  c: word; i, j: integer;
  f: array [0..m] of integer;
  StNicel: array [0..b - 1] of integer;
begin
  { Za vsako mono vrednost pretejmo, kolikokrat se pojavlja. }
  for c := 0 to m do f[c] := 0;
  for i := 0 to n - 1 do f[Blok[i]] := f[Blok[i]] + 1;
  { Pretejmo, koliko je v bloku niel na vsakem bitu. }
  for j := 0 to b - 1 do StNicel[j] := 0;
  for c := 0 to m do if f[c] > 0 then begin
    for j := 0 to b - 1 do 
      if not Odd(c shr j) then StNicel[j] := StNicel[j] + f[c];
  end; {for c}
  { V c prigimo tiste bite, kjer je enic ve kot niel. }
  c := 0;
  for j := 0 to b - 1 do
    if StNicel[j] < n - StNicel[j] then
      c := c or (1 shl j);
  NajvecNicel3b := c;
end; {NajvecNicel3b}


function Ura: double;
var 
  ct, et, kt, ut: TFILETIME;
  ul: ULARGE_INTEGER;
begin
  GetProcessTimes(GetCurrentProcess, @ct, @et, @kt, @ut);
  ul.LowPart := ut.dwLowDateTime;
  ul.HighPart := ut.dwHighDateTime;
  Ura := ul.QuadPart / 10000000.0;
end; {Ura}

type FunkcijaT = function (const Blok: BlokT): word;

procedure InitBlok(var Blok: BlokT);
var i: integer;
begin
  for i := 0 to n - 1 do
    Blok[i] := (Random(65536) or (Random(65536) shl 16)) and m;
end; {InitBlok}

procedure Test(f: FunkcijaT; Ime: string);
var Blok: ^BlokT; tm, tmStart: double; nIter: integer; c: word;
begin
  New(Blok);
  nIter := 0; tm := 0;
  RandSeed := 123;
  repeat
    InitBlok(Blok^);
    tmStart := Ura;
    c := f(Blok^);
    tm := tm + (Ura - tmStart);
    nIter := nIter + 1;
    if nIter = 1 then Write(Ime, ': c = ', c);
  until tm > 1.0;
  WriteLn('; ', tm:0:2, ' s for ', nIter, ' iterations = ', (tm / nIter):0:6, ' s/iter');
  Dispose(Blok);
end; {Blok}

begin {DrustvoLjubiteljevNicel}
  Test(@NajvecNicel1, 'NajvecNicel1');
  Test(@NajvecNicel1b, 'NajvecNicel1b');
  Test(@NajvecNicel2, 'NajvecNicel2');
  Test(@NajvecNicel2b, 'NajvecNicel2b');
  Test(@NajvecNicel3, 'NajvecNicel3');
  Test(@NajvecNicel3b, 'NajvecNicel3b');
end. {DrustvoLjubiteljevNicel}
